****************************************************************************************************
*                                           PROGRAM OVERVIEW
****************************************************************************************************
*
* PROGRAM: ms_logcheck.sas  
*
* Created (mm/dd/yyyy): 09/18/2014
* Last modified: 09/24/15 
* Version: 2.0
*
*--------------------------------------------------------------------------------------------------
* PURPOSE:
*  This log checker tool can be used to consistently and efficiently scan resultant SAS logs for any
*  concerning lines that may indicate an issue with program execution. 
*
*
* PARAMETERS:
*   InputDir= The directory (full path) containing one or more logs to be scanned.
*   InputFile= The name of the specific log file(s) to be scanned, with the ".log" suffix. IF SPECIFYING MORE THAN ONE LOG FILE, SEPARATE WITH "^". Optional.
*   OutputDir= The directory (full path) to which the output report should be written.
*   Keywords= The format of the resultant output report.  Optional.  Default=pdf.  Allowable values: pdf, html, rtf.
*   ContextLines= The number of lines above and below the line of interest that will be reported along with the line of interest.
*                 Optional. Default=0. Maximum=10 (any number > 10 will be capped at 10).
*   ReportType= Additional, user-specified keywords for which the tool will scan and report.  Optional. 
*  
*--------------------------------------------------------------------------------------------------
* CONTACT INFO: 
*  Sentinel Coordinating Center
*  info@mini-sentinel.org
*
*--------------------------------------------------------------------------------------------------
*  CHANGE LOG: 
*
*   Version   Date       Initials      Comment (reference external documentation when available)
*   -------   --------   --------   ---------------------------------------------------------------
*    1.0      02/08/2015  RdeF     
*    2.0      09/18/2015  NN         Added code in lines 101-200 to update functionality--can now 
*							         specify multiple logs versus just one
*
***************************************************************************************************;

options symbolgen mprint mlogic;

%MACRO MS_LOGCHECK(INPUTDIR=, INPUTFILE=, OUTPUTDIR=, KEYWORDS=, CONTEXTLINES=, REPORTTYPE=);

/*---------------------------------------------------------------------------*/  
/*                     MS_LOGCHECK Version 2.0                               */
/*---------------------------------------------------------------------------*/

%put =====> MACRO CALLED: ms_logcheck v2.0;

/*-- Capture run date and time--*/
data _null_;
   call symput('CURDATE',strip(put(date(),worddate.) || ' @' || put(time(),hhmm.)));
run;

%LET NLOGS=0;
%LET FILNAMES=NONE;

%MACRO LOCATE_LOGS;
%IF "&INPUTFILE" = "%str()" %THEN %DO;
%LET NLOGS=0;

/*--Search directory for log files if no log file specified--*/
/*--If no log specified, code reads through all logs--*/
data _logfiles;
   keep fname;
   length filref $8 fname $80; 
   rc = filename(filref, "&INPUTDIR"); 
   if rc = 0 then do; 
      dir_id = dopen(filref); 
      rc = filename(filref);  
   end; 
   else do; 
      length msg $200.; 
      msg = sysmsg(); 
      put msg=; 
      dir_id = .; 
      putlog 'ERR' 'OR: Unable to open directory.';
      abort cancel;
   end;
   n_members = dnum(dir_id);
   num_logs=0;
   do i = 1 to n_members; 
      fname = dread(dir_id, i); 
      fid = mopen(dir_id, fname); 
      if fid > 0 and reverse(compress(fname)) =: 'gol.' then do;
         num_logs=num_logs+1;
         output;
      end;
   end;
   rc = dclose(dir_id);
   call symputx('NLOGS',strip(put(num_logs,5.)));
run;
proc sql noprint;
select strip(fname) into :FILNAMES separated by '^'
   from _logfiles;
quit;

%END; 

/********************************************************************************************************/
/********* Updated functionality to include multiple (but not all) logs in directory 09/10/15 NN*********/
/********************************************************************************************************/

%ELSE %IF %sysfunc(countc( "&INPUTFILE", '^'  )) ge 1 %THEN %DO;
/*--Save log file if multiple log files are specifed--*/
	%LET NLOGS=%sysfunc(countc( "&INPUTFILE", '^'  ))+1;
	%LET FILNAMES= &INPUTFILE;
%END;


%ELSE %DO;
/*--Save log file if ONE specified log file--*/
   %LET NLOGS=1;
   %LET FILNAMES = &INPUTFILE/*..log*/;
%END; 

/********************************************************************************************************/
/********************************************************************************************************/
/********************************************************************************************************/




%MEND LOCATE_LOGS;

%LOCATE_LOGS;

%PUT N_LOGS = &NLOGS;
%PUT FILNAMES = &FILNAMES;

%IF %EVAL(&NLOGS = 0) %THEN %DO;
   %PUT ________________________________________________________________________________________________;
   %PUT ERROR: No logs located in specified folder: "&INPUTDIR";
   %ABORT cancel;
%END;

/*--Phrases to look for in notes to indicate an error--internal to this program (not user special)--*/
%LET NOTES = stopped|converted|uninitialized|Division|Invalid|invalid|Missing|W.D format|repeats of BY values|;
%LET NOTES = &NOTES.Mathematical|MERGE statement|Character values have disabled|No observations|Groups are not| 0 observations;
%LET NUSER = &KEYWORDS;

%PUT NUSER = &NUSER;

/*--Date will be in footnote, not title--*/
options nodate;

/*--Format for labeling issues--*/
proc format;
value errfmt
   4='ERROR'
   3='WARNING'
   2='NOTE'
   1='SPECIAL'
   other=' ';
quit;

/*--Convert macro CONTEXTLINES from blank to zero and cap at 10--*/
%IF "&CONTEXTLINES" = "%STR()" %THEN %LET CONTEXTLINES=0;
%IF %EVAL(&CONTEXTLINES > 10) %THEN %LET CONTEXTLINES=10;

/*--Convert macro REPORTTYPE to [default] pdf if left blank--*/
%IF "&REPORTTYPE" = "%STR()" %THEN %LET REPORTTYPE=pdf;

/*--Loop through all logs looking for errors/warnings/notes/specials and reporting--*/
%DO FS = 1 %TO &NLOGS;
   %LET FN = %SCAN(&FILNAMES,&FS,'^');
   %LET OUTREPORT = &OUTPUTDIR.%SUBSTR(&FN,1,%INDEX(&FN,.log)-1)_logerr;

   %PUT OUTREPORT = &OUTREPORT;

/*--Read in log file--*/
data _input_log;
   infile "&INPUTDIR.&FN";
   lineno+1;
   input;
   log_line = _infile_;
run;

/*--Search for issues in the log file--*/
data _pull_errors(drop=num_:) _summary_nos(keep=num_:);
      set _input_log end=eof;
   retain rid_e1 rid_e2 rid_w1 rid_w2 rid_n1 rid_n2 rid_us;
   retain num_error num_warning num_note num_special 0;
   length err_level 3;
   err_level=0;
   /*--Compile regular expressions--*/
   if lineno=1 then do; 
      rid_e1 = prxparse('/^ERROR:/');
      rid_e2 = prxparse('/^ERROR [0-9]/');
      rid_w1 = prxparse('/^WARNING:/');
      rid_w2 = prxparse('/^WARNING [0-9]/');
      rid_n1 = prxparse('/^NOTE:/');
      rid_n2 = prxparse("/(&NOTES)/");
      rid_us = prxparse("/(&KEYWORDS)/");
   end;

   drop rid_:;

   /*--search log line for ERROR, WARNING, NOTE with specified phrases--*/
   /*                   or user specied (special) NOTE phrases          */
   if prxmatch(rid_e1,log_line) or prxmatch(rid_e2,log_line) then do;
      err_level=4;
      num_error=num_error+1;
      output _pull_errors;
   end; else
   if prxmatch(rid_w1,log_line) or prxmatch(rid_w2,log_line) then do;
      err_level=3;
      num_warning=num_warning+1;
      output _pull_errors;
   end; else 
   if (prxmatch(rid_n1,log_line) and prxmatch(rid_n2,log_line)) then do;
      err_level=2;
      num_note=num_note+1;
      output _pull_errors;
   end; 

   %IF "&NUSER" NE "%str()" %THEN %DO;

   else
   if prxmatch(rid_us,log_line) then do;
      err_level=1;
      num_special=num_special+1;
      output _pull_errors;
   end;

   %END;

   if eof then do;
   /*--Free memory used by regular expression--*/
      call prxfree(rid_e1);
      call prxfree(rid_e2);
      call prxfree(rid_w1);
      call prxfree(rid_w2);
      call prxfree(rid_n1);
      call prxfree(rid_n2);
      call prxfree(rid_us);
      output _summary_nos;
   end;
run;

/*--Prepare summary report numbers--*/
proc transpose data=_summary_nos out=_summary_nos prefix=count;
run;   
data _summary_nos; set _summary_nos;
   keep count1 num_no;
   rename count1=count;

   num_no = 5-_N_;
run;

/*--Create list of rows to report with context lines included in detailed report--*/
data _expanded_list(rename=(i=lineno new_elev=err_level)); set _pull_errors;
   keep i new_elev;
   list_flag=0;
   l_start = lineno-&CONTEXTLINES;
   l_finsh = lineno+&CONTEXTLINES;
   do i = l_start to l_finsh;
      if i = lineno then new_elev = err_level; else new_elev=0;
      output;
   end;
run;

/*--Prepare list (proper order, no duplicates, merge with log data) of issues for detail reporting--*/
proc sort data=_expanded_list; by lineno descending err_level;
run;
data _expanded_list; set _expanded_list; by lineno;
   if first.lineno;
run;
data _final_list; merge _input_log _expanded_list(in=inxp); by lineno;
   if inxp;
run;

/*--Set up output reports--*/
ods listing close;
ods &REPORTTYPE FILE="&OUTREPORT..&REPORTTYPE";

/*--Produce summary report--*/
proc report data=_summary_nos nowd;
   column num_no count tmp;
   define num_no / display 'Error Level' center format=errfmt.;
   define count / display 'Number of Issues' center format=comma7.;
   define tmp / computed noprint;

   compute tmp;
      if num_no = 4 and count > 0 then
         call define
           (_row_,"style","style=[background=RED]");
      else
      if num_no = 3 and count > 0 then
         call define
            (_row_,"style","style=[background=ORANGE]");
      else   
      if num_no = 2 and count > 0 then
         call define
            (_row_,"style","style=[background=YELLOW]");
      else
      if num_no = 1 and count > 0 then
         call define
            (_row_,"style","style=[background=POWDERBLUE]");
   endcomp;

   title "MSOC LOG ERROR CHECKER (v2.0)";
   title2 'Summary Report';
   footnote "Error Report: &OUTREPORT..&REPORTTYPE";
   footnote2 "Run Date/Time: &CURDATE.";
run;

/*--Produce detailed report--*/
proc report data=_final_list nowd;
   where log_line ne ' ';
   column lineno err_level log_line tmp;
   define lineno / order order=internal 'Line Number' center;
   define err_level / display 'Error Level' center format=errfmt.;
   define log_line / display 'Log Line';
   define tmp / computed noprint;

   compute tmp;
      if err_level = 4 then
         call define
            (_row_,"style","style=[background=RED]");
      else   
      if err_level = 3 then
         call define
            (_row_,"style","style=[background=ORANGE]");
      else   
      if err_level = 2 then
         call define
            (_row_,"style","style=[background=YELLOW]");
      else
      if err_level = 1 then
         call define
            (_row_,"style","style=[background=POWDERBLUE]");
   endcomp;
   title "MSOC LOG ERROR CHECKER (v2.0)";
   title2 'Detail Report';
   footnote "Error Report: &OUTREPORT..&REPORTTYPE";
   footnote2 "Run Date/Time: &CURDATE.";

   %IF %EVAL(&CONTEXTLINES = 0) %THEN %DO;

   footnote3 height=2 italic "(No extra lines surrounding the highlighted issues are displayed)";

   %END; %ELSE %DO;

   footnote3 height=2 italic
   "(&CONTEXTLINES lines surrounding the highlighted issues are displayed unless colliding with adjacent issues.  Blank lines are not displayed.)";

   %END;

run ;   

/*--Close output reports--*/
ODS &REPORTTYPE CLOSE;
ods listing;
run;

/*--Delete temp data sets--*/
proc datasets lib=work;
   delete _expanded_list _final_list _input_log _pull_errors _summary_nos

   %IF %SYSFUNC(exist(_logfiles)) %THEN %DO;

   _logfiles 

   %END;

   ;
quit;

%END;

%put NOTE: ******** END OF MACRO: ms_logcheck v2.0 ********;


%MEND MS_LOGCHECK;


